from viggocore.celery import celery, decide_on_run
from viggocore.subsystem.image.handler import ImageHandler


@decide_on_run
@celery.task
def process_image(folder: str, filename: str) -> None:
    handler = ImageHandler()
    handler.process(folder, filename)



import flask

from viggocore.common import exception, utils
from viggocore.common.subsystem import controller


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def _get_user_from_token(self):
        if flask.has_request_context():
            token_id = flask.request.headers.get('token')
            if token_id is not None:
                self.token = self.manager.api.tokens().get(id=token_id)
                return self.token.user_id
        return None

    # TODO descobrir alguma forma de reaproveitar o list do viggocore
    def list(self):
        filters = self._filters_parse()
        filters = self._filters_cleanup(filters)

        try:
            user_id = self._get_user_from_token()
            filters = self._parse_list_options(filters)

            if user_id is not None:
                filters['user_id'] = user_id
            else:
                return flask.Response("user_id not found",
                                      status=403)

            timeline_events = self.manager.list(**filters)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        timeline_events_dict = self._entities_to_dict(
                timeline_events, self._get_include_dicts())
        response = {self.collection_wrap: timeline_events_dict}

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")



import os

from PIL import Image

from viggocore.subsystem.image.resource import Image as ImageViggocore, \
    QualityImage


class ImageHandler:

    def __init__(self, sizes=None):
        self.target_extension = '.jpg'
        if sizes is None:
            self.sizes = {
                QualityImage.min.value: (170, 150),
                QualityImage.med.value: (350, 310),
                QualityImage.max.value: (1500, 1300)
            }
        else:
            self.sizes = sizes

    def process(self, folder, filename):
        try:
            image_path = os.path.join(folder, filename)
            with Image.open(image_path) as image:
                image.load()
                normalized = self.normalize(image)
                file, extension = os.path.splitext(filename)
                self.create_thumbnails(normalized, os.path.join(folder, file))
        except IOError:
            raise

    def _remove_transparency(self, image):
        fill_color = (255, 255, 255)
        image = image.convert('RGBA')
        try:
            mask = image.split()[-1]
        except IndexError:
            mask = None
        background = Image.new('RGBA', image.size, fill_color)
        background.paste(image, mask=mask)
        image = background
        return image

    def normalize(self, image):
        if image.mode in ['RGBA', 'P', 'LA']:
            image = self._remove_transparency(image)
        if image.mode != 'RGB':
            image = image.convert('RGB')
        return image

    def create_thumbnails(self, image, file):
        for size in self.sizes:
            name = '{}.{}{}'.format(file, size, self.target_extension)
            self.create_thumbnail(image, self.sizes[size], name)

    def create_thumbnail(self, image, size, name):
        width, height = size
        if image.width < width:
            width = image.width
        if image.height < height:
            height = image.height
        thumb = image.copy()
        thumb.thumbnail((width, height))
        thumb.save(name, 'JPEG')

    @staticmethod
    def verify_size_resolution_image(folder, filename):
        messageError = None
        try:
            image_path = os.path.join(folder, filename)
            with Image.open(image_path) as img:
                if img.height > ImageViggocore.MAX_ALLOWED_SIZE[0] or \
                   img.width > ImageViggocore.MAX_ALLOWED_SIZE[1]:
                    messageError = 'Image must be less than {}px and less\
                        than {}px high!'.format(
                            ImageViggocore.MAX_ALLOWED_SIZE[0],
                            ImageViggocore.MAX_ALLOWED_SIZE[1])
        except Exception:
            messageError = 'ERROR: image resolution check failed'
        finally:
            return messageError



from enum import Enum

import sqlalchemy
from viggocore.common.subsystem import entity
from viggocore.database import db
from sqlalchemy import UniqueConstraint


class RoleDataViewType(Enum):
    MULTI_DOMAIN = {
        'id': 0,
        'description': 'tem acesso aos dados de mais de um domínio.'}
    DOMAIN = {
        'id': 1,
        'description': 'tem acesso aos dados de apenas um domínio.'}


class Role(entity.Entity, db.Model):

    USER = 'User'
    SYSADMIN = 'Sysadmin'
    ADMIN = 'Admin'
    SUPORTE = 'Suporte'

    attributes = ['name', 'data_view']
    attributes += entity.Entity.attributes

    name = db.Column(db.String(80), nullable=False)
    data_view = db.Column(sqlalchemy.Enum(RoleDataViewType), nullable=False,
                          default=RoleDataViewType.DOMAIN,
                          server_default='DOMAIN')

    __table_args__ = (
        UniqueConstraint('name', name='role_name_uk'),)

    def __init__(self, id, name, data_view,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.name = name
        self.data_view = data_view



import os
import shutil
from pathlib import Path

from viggocore.common.subsystem import operation
from viggocore.subsystem.file import manager
from viggocore.subsystem.image import tasks
from viggocore.subsystem.image.resource import QualityImage
from viggocore.subsystem.image.handler import ImageHandler
from viggocore.common import exception


class Create(manager.Create):

    def do(self, session, **kwargs):
        super().do(session)

        # TODO (araujobd)  check a better way to improve this
        # only way to validate resolution of image was after
        # save temporary file so if exceeds should trigger rollback
        folder = self.manager.get_upload_folder(self.entity,
                                                self.entity.domain_id)
        error_message = ImageHandler.verify_size_resolution_image(
            folder, self.entity.filename)
        if error_message is not None:
            shutil.rmtree(folder)
            raise exception.BadRequest(error_message)

        return self.entity

    def post(self):
        tasks.process_image(self.upload_folder, self.entity.filename)


class Get(operation.Get):

    def pre(self, session, id, **kwargs):
        self.quality = kwargs.pop('quality', QualityImage.min)

        return super().pre(id=id, session=session)

    def do(self, session, **kwargs):
        file = super().do(session=session, **kwargs)

        folder = self.manager.get_upload_folder(file, file.domain_id)
        filename = file.filename_with_quality(self.quality)

        existingFile = Path(f'{folder}/{filename}')
        if existingFile.is_file() is False and self.quality is not None:
            filename = file.filename_with_quality(None)
            existingFile = Path(f'{folder}/{filename}')

        if existingFile.is_file() is False:
            raise exception.ViggoCoreException('File not found!')
        else:
            return folder, filename


class Delete(operation.Delete):

    def post(self):
        # TODO(fdoliveira) Put this in worker
        folder = self.manager.get_upload_folder(self.entity,
                                                self.entity.domain_id)
        shutil.rmtree(folder)


class Manager(manager.Manager):
    ALLOWED_EXTENSIONS = ['jpg', 'jpeg', 'png', 'gif']

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.get = Get(self)
        self.delete = Delete(self)

    def get_upload_folder(self, entity, domain_id):
        base_folder = self._get_base_folder()
        entity_name = type(entity).__name__
        folder = os.path.join(base_folder,
                              entity_name,
                              entity.type_image,
                              self.OPTIONAL_FOLDER,
                              domain_id,
                              entity.id)
        if not os.path.exists(folder):
            os.makedirs(folder)
        return folder



from viggocore.common.subsystem.apihandler import Api
import uuid

from typing import Dict, List

from viggocore.common.subsystem import Subsystem
from viggocore.subsystem.route.resource import Route


class BootstrapRoutes(object):

    def __init__(self, subsystems: Dict[str, Subsystem], api: Api) -> None:
        self.route_manager = api.routes()
        self.subsystems_routes = self._get_routes_subsystems(subsystems)

    def _get_routes_subsystems(self,
                               subsystems: Dict[str, Subsystem]) -> List[Dict]:
        routes = [route for subsystem in subsystems.values()
                  for route in subsystem.router.routes]
        return routes

    def execute(self) -> List[Route]:
        routes_db = self._get_routes_db()
        # TODO inactivate old_routes
        (new_routes, old_routes) = self._get_diff_routes(
            self.subsystems_routes, routes_db)

        new_routes_db = self._get_create_routes(new_routes)
        self._save_routes(new_routes_db)

        return new_routes_db + routes_db

    def _get_routes_db(self) -> List[Route]:
        routes = self.route_manager.list()
        return routes

    def _get_diff_routes(self, routes_app: List[Dict], routes_db: List[Route]):
        new_routes, old_routes = [], []

        routes_db_tuples = [(route.url, route.method) for route in routes_db]
        routes_app_tuples = [(route['url'], route['method'])
                             for route in routes_app]

        for route in routes_app:
            if (route['url'], route['method']) not in routes_db_tuples:
                new_routes.append(route)

        for route in routes_db:
            if (route.url, route.method) not in routes_app_tuples:
                old_routes.append(route)

        return (new_routes, old_routes)

    def _get_create_routes(self, routes: List[Dict]) -> List[Route]:
        new_routes_db = []
        for route in routes:
            name = route['action']
            url = route['url']
            method = route['method']
            bypass = route.get('bypass', False)
            new_route = Route(id=uuid.uuid4().hex,
                              name=name,
                              url=url,
                              method=method,
                              bypass=bypass)
            new_routes_db.append(new_route)
        return new_routes_db

    def _save_routes(self, routes: List[Route]):
        if routes:
            self.route_manager.create_routes(routes=routes)



from viggocore.common.subsystem import manager, entity
from sqlalchemy import func, and_, or_
from datetime import datetime as datetime1
import datetime as datetime2


class CommonManager(manager.Manager):

    def __init__(self, driver):
        super(CommonManager, self).__init__(driver)

    def apply_filters_includes(self, query, dict_compare, **kwargs):
        for id, resource in dict_compare.items():
            for k, v in kwargs.items():
                if id in k and hasattr(resource, k.split('.')[-1]):
                    k = k.split('.')[-1]
                    isinstance_aux = isinstance(v, str)

                    if k == 'tag':
                        # TODO(JorgeSilva): definir o caractere para split
                        values = v
                        if len(v) > 0 and v[0] == '#':
                            values = v[1:]
                        values = values.split(',')
                        filter_tags = []
                        for value in values:
                            filter_tags.append(
                                getattr(resource, k)
                                .like('%#'+str(value)+' %'))
                        query = query.filter(or_(*filter_tags))
                    elif isinstance_aux and self.__isdate(v):
                        day, next_day = self.__get_day_and_next_day(v)
                        query = query.filter(
                            and_(
                                or_(getattr(resource, k) < next_day,
                                    getattr(resource, k) == None),  # noqa: E711
                                or_(getattr(resource, k) >= day,
                                    getattr(resource, k) == None)))  # noqa: E711 E501
                    elif isinstance_aux and '%' in v:
                        normalize = func.viggocore_normalize
                        query = query.filter(normalize
                                             (getattr(resource, k))
                                             .ilike(normalize(v)))
                    else:
                        query = query.filter(getattr(resource, k) == v)

        return query

    def apply_filters(self, query, resource, **kwargs):
        for k, v in kwargs.items():
            if '.' not in k and hasattr(resource, k):
                isinstance_aux = isinstance(v, str)

                if k == 'tag':
                    # TODO(JorgeSilva): definir o caractere para split
                    values = v
                    if len(v) > 0 and v[0] == '#':
                        values = v[1:]
                    values = values.split(',')
                    filter_tags = []
                    for value in values:
                        filter_tags.append(
                            getattr(resource, k)
                            .like('%#'+str(value)+' %'))
                    query = query.filter(or_(*filter_tags))
                elif isinstance_aux and self.__isdate(v):
                    day, next_day = self.__get_day_and_next_day(v)
                    query = query.filter(
                        and_(
                             or_(getattr(resource, k) < next_day,
                                 getattr(resource, k) == None),  # noqa: E711
                             or_(getattr(resource, k) >= day,
                                 getattr(resource, k) == None)))  # noqa: E711
                elif isinstance_aux and '%' in v:
                    normalize = func.viggocore_normalize
                    query = query.filter(normalize(getattr(resource, k))
                                         .ilike(normalize('%'+str(v))))
                elif isinstance(v, str) and '':
                    query = query.filter(
                    )
                else:
                    query = query.filter(getattr(resource, k) == v)

        return query

    def with_pagination(self, **kwargs):
        require_pagination = kwargs.get('require_pagination', False)
        page = kwargs.get('page', None)
        page_size = kwargs.get('page_size', None)

        if (page and page_size is not None) and require_pagination is True:
            return True
        return False

    def __isdate(self, data, format="%Y-%m-%d"):
        res = True
        try:
            res = bool(datetime1.strptime(data, format))
        except ValueError:
            res = False
        return res

    def __get_day_and_next_day(self, data, format="%Y-%m-%d"):
        day = datetime1.strptime(data, format)
        next_day = day + datetime2.timedelta(days=1)
        return (day, next_day)

    def apply_filter_de_ate(self, resource, query, de, ate):
        inicio = datetime1.strptime(de, entity.DATE_FMT)
        fim = datetime1.strptime(ate, entity.DATE_FMT) +\
            datetime2.timedelta(days=1)
        return query.filter(
            and_(resource.created_at > inicio, resource.created_at < fim))



from viggocore.common import subsystem
from viggocore.subsystem.application import resource, manager, controller, \
    router

subsystem = subsystem.Subsystem(resource=resource.Application,
                                manager=manager.Manager,
                                controller=controller.Controller,
                                router=router.Router)



import json
import uuid
import unittest
import testtools

from viggocore.tests.functional import test_base


class GrantTestCase(test_base.CreateTest,
                    test_base.RetrieveTest,
                    test_base.DeleteTest,
                    testtools.TestCase):

    def load_fixtures(self):
        domain = self.post(
            resource_ref={'name': uuid.uuid4().hex, 'active': True},
            resource_name='domain',
            collection_name='domains')



        self.domain_id = domain['id']

    @property
    def resource_name(self):
        return 'grant'

    @property
    def required_attributes(self):
        return ['user_id', 'role_id']

    @property
    def unique_attributes(self):
        return [('user_id', 'role_id')]

    def new_resource_ref(self):
        user = self.post(
            resource_ref={'name': uuid.uuid4().hex,
                          'domain_id': self.domain_id,
                          'email': uuid.uuid4().hex,
                          'password': uuid.uuid4().hex,
                          'active': True},
            resource_name='user',
            collection_name='users')

        role = self.post(
            resource_ref={'name': uuid.uuid4().hex,
                          'domain_id': self.domain_id,
                          'active': True},
            resource_name='role',
            collection_name='roles')

        return {'user_id': user['id'],
                'role_id': role['id']}


if __name__ == '__main__':
    unittest.main()
